
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/wait.h>
#include <string.h>
#include "project.h"


void reverseArray(int arr[], int size) { //An array to reverse the numbers
        int left = 0;
        int right = size - 1;
        while (left < right) {
                int temp = arr[left];
                arr[left] = arr[right];
                arr[right] = temp;
                left++;
                right--;
        }
}

int main(int argc, char *argv[]) {
const char *name = argv[0]; //Getting name of excutable for error checking
int opt; 
const char *inputFilename = "input.dat";  //Defaults for user not specifying the name
const char *outputFilename = "output.dat";

while (( opt = getopt (argc, argv, "hio")) != -1){
        switch (opt){ //Opt to parse command line options
        case 'h':
            printH();  //Calls functon to print help statement and menu options
            return (EXIT_SUCCESS);
        case 'i':
            if (optarg != NULL) {
                inputFilename = optarg;
            } else {
                inputFilename = "input.dat";
            }
            break;
        case 'o':
            if (optarg != NULL){
                outputFilename = optarg;
            } else {
                outputFilename = "output.dat";
            }
            break;
        default: //Error checking
            fprintf (stderr, "%s: Error: Invalid option, please refer to the help menu\n", name);
            return(EXIT_FAILURE);
        }

}

int shm_id = shmget(IPC_PRIVATE, sizeof(long) * 10, IPC_CREAT | 0666); //Allocating shared memory
if (shm_id < 0){ //Error checking
        perror("Error: Shared memory allocation failed\n");
        fprintf(stderr, "%s", name);
        exit(1);
}

long *filePosition = shmat(shm_id, NULL, 0); //Pointer for shared memory
if (filePosition == (void *) -1){
        perror("Error: Error attaching shared memory\n");
        fprintf(stderr, "%s", name);
        exit(1);
}

FILE *inputFile = fopen(inputFilename, "r"); //Opening read and write for the files
FILE *outputFile = fopen(outputFilename, "w");

if (inputFile == NULL){ //Error checking if files do not open correctly
        fprintf(stderr, "%s: Error: Failed to open input file", name);
        return(EXIT_FAILURE);
}
if (outputFile == NULL){
        fprintf(stderr, "%s: Error: Failed to open output file", name);
        return(EXIT_FAILURE);
}

*filePosition = ftell(inputFile); //Pointer to file position

int tasks;
fscanf(inputFile, "%d", &tasks); //Scanning in the number in order to fork

pid_t childP[tasks];
pid_t parentPID = getpid();
int totalC = 0;
int line = 1;


for (int t = 0; t < tasks; t++){
        pid_t childPID = fork(); //Fork process begins
        if (childPID == 0){
                long position = *filePosition;
                fseek(inputFile, position, SEEK_SET);
                for(int i = 0; i < line; i++){ //Loop to skip the processes
                        if(fscanf(inputFile, "%*[^\n]\n") == EOF){
                                exit(1);
                        };
                }

                int numInt;
                fscanf(inputFile, "%d", &numInt); //Scanning in next number
                int integers[numInt];

                for(int j = 0; j < numInt; j++) { //Scanning in final list of numbers
                        fscanf(inputFile, "%d", &integers[j]); 
                }

                fprintf(outputFile, "%d", getpid()); //Printing the PIDS to the output file
                reverseArray(integers, numInt);  //Reversing the numbers

                 for(int j = 0; j < numInt; j++){ //Printing the reversed numbers
                        fprintf(outputFile, " %d", integers[j]);
                }

                fprintf(outputFile, "\n");
                
                fclose(outputFile);
                fclose(inputFile);
                } else if (childPID < 0) { //Error checking
                        fprintf(stderr, "%s: Error: Fork did not work", name);
                        exit(1);
                } else {
                        childP[t] = childPID;
                }
                totalC = t + 1; //Updating and incrementing to the next file lines
                line += 2;
        }
        
        sleep(3); //Waiting for the process forking to complete

        fprintf(outputFile, "\nAll children were: "); //Outputting the children
        for(int i = 0; i < totalC; i++){
                fprintf(outputFile, "%d", childP[i]);
        }

        fprintf(outputFile, "\nParent PID: %d\n", parentPID);  //Outputting the parent PID

        fclose(outputFile);

        shmdt(filePosition); //Freeing pointers
        shmctl(shm_id, IPC_RMID, NULL);

        return 0;
}       

